/**
 * 
 */
package org.gcube.portlets.user.tdtemplate.server.converter;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Calendar;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.gcube.data.analysis.tabulardata.commons.templates.model.Template;
import org.gcube.data.analysis.tabulardata.commons.templates.model.TemplateCategory;
import org.gcube.data.analysis.tabulardata.commons.templates.model.actions.finals.DuplicateBehaviour;
import org.gcube.data.analysis.tabulardata.commons.templates.model.columns.ColumnCategory;
import org.gcube.data.analysis.tabulardata.commons.templates.model.columns.ColumnDescription;
import org.gcube.data.analysis.tabulardata.commons.templates.model.columns.TemplateColumn;
import org.gcube.data.analysis.tabulardata.commons.utils.DimensionReference;
import org.gcube.data.analysis.tabulardata.commons.utils.FormatReference;
import org.gcube.data.analysis.tabulardata.commons.utils.Licence;
import org.gcube.data.analysis.tabulardata.commons.utils.LocaleReference;
import org.gcube.data.analysis.tabulardata.commons.utils.TimeDimensionReference;
import org.gcube.data.analysis.tabulardata.commons.webservice.types.OnRowErrorAction;
import org.gcube.data.analysis.tabulardata.commons.webservice.types.TemplateDescription;
import org.gcube.data.analysis.tabulardata.expression.Expression;
import org.gcube.data.analysis.tabulardata.model.DataTypeFormats;
import org.gcube.data.analysis.tabulardata.model.ValueFormat;
import org.gcube.data.analysis.tabulardata.model.datatype.DataType;
import org.gcube.data.analysis.tabulardata.model.metadata.Locales;
import org.gcube.data.analysis.tabulardata.model.metadata.common.TableDescriptorMetadata;
import org.gcube.data.analysis.tabulardata.model.table.Table;
import org.gcube.data.analysis.tabulardata.model.table.TableId;
import org.gcube.data.analysis.tabulardata.model.time.PeriodType;
import org.gcube.data.analysis.tabulardata.service.tabular.metadata.AgencyMetadata;
import org.gcube.data.analysis.tabulardata.service.tabular.metadata.DescriptionMetadata;
import org.gcube.data.analysis.tabulardata.service.tabular.metadata.LicenceMetadata;
import org.gcube.data.analysis.tabulardata.service.tabular.metadata.NameMetadata;
import org.gcube.data.analysis.tabulardata.service.tabular.metadata.RightsMetadata;
import org.gcube.data.analysis.tabulardata.service.tabular.metadata.TabularResourceMetadata;
import org.gcube.data.analysis.tabulardata.service.tabular.metadata.ValidSinceMetadata;
import org.gcube.data.analysis.tabulardata.service.tabular.metadata.ValidUntilMetadata;
import org.gcube.portlets.user.td.gwtservice.shared.tr.ColumnData;
import org.gcube.portlets.user.td.widgetcommonevent.shared.TRId;
import org.gcube.portlets.user.td.widgetcommonevent.shared.tr.TabResourceType;
import org.gcube.portlets.user.tdtemplate.server.TemplateUpdaterForDescription;
import org.gcube.portlets.user.tdtemplate.server.service.ExpressionEvaluatorInstance;
import org.gcube.portlets.user.tdtemplate.server.service.TemplateService;
import org.gcube.portlets.user.tdtemplate.server.session.CacheServerExpressions;
import org.gcube.portlets.user.tdtemplate.server.validator.ColumnCategoryConstraint;
import org.gcube.portlets.user.tdtemplate.server.validator.service.ColumnCategoryTemplateValidator;
import org.gcube.portlets.user.tdtemplate.shared.SPECIAL_CATEGORY_TYPE;
import org.gcube.portlets.user.tdtemplate.shared.TdBehaviourModel;
import org.gcube.portlets.user.tdtemplate.shared.TdColumnDefinition;
import org.gcube.portlets.user.tdtemplate.shared.TdFlowModel;
import org.gcube.portlets.user.tdtemplate.shared.TdLicenceModel;
import org.gcube.portlets.user.tdtemplate.shared.TdTColumnCategory;
import org.gcube.portlets.user.tdtemplate.shared.TdTDataType;
import org.gcube.portlets.user.tdtemplate.shared.TdTFormatReference;
import org.gcube.portlets.user.tdtemplate.shared.TdTFormatReferenceIndexer;
import org.gcube.portlets.user.tdtemplate.shared.TdTTemplateType;
import org.gcube.portlets.user.tdtemplate.shared.TdTTimePeriod;
import org.gcube.portlets.user.tdtemplate.shared.TdTemplateDefinition;
import org.gcube.portlets.user.tdtemplate.shared.TdTemplateUpdater;
import org.gcube.portlets.user.tdtemplate.shared.TemplateExpression;
import org.gcube.portlets.user.tdtemplate.shared.util.CutStringUtil;
import org.gcube.portlets.user.tdtemplate.shared.validator.ViolationDescription;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * @author Francesco Mangiacrapa francesco.mangiacrapa@isti.cnr.it
 * @Jan 17, 2014
 *
 */
public class ConverterToTdTemplateModel {
	
	
	public static Logger logger = LoggerFactory.getLogger(ConverterToTdTemplateModel.class);
	
	/**
	 * 
	 * @return
	 */
	public static List<TdTTemplateType> getTdTTemplateTypeFromTemplateCategoryValues(){
		
		List<TdTTemplateType> templates = new ArrayList<TdTTemplateType>(TemplateCategory.values().length);
		for (TemplateCategory template : TemplateCategory.values()) {
			
			List<ViolationDescription> constraint = getTemplateCategoryConstraint(template);
			templates.add(new TdTTemplateType(template.name(), template.name(), constraint));
		}
		
		logger.info("Returning "+templates.size()+" template types");
		
		return templates;
		
	}
	
	public static List<ViolationDescription> getTemplateCategoryConstraint(TemplateCategory cat) {

		List<ViolationDescription> violations  = new ArrayList<ViolationDescription>();
		
		ColumnCategoryTemplateValidator templateValidator = null;
		 
		switch (cat) {
			case CODELIST:
				templateValidator = new ColumnCategoryTemplateValidator(TemplateCategory.CODELIST);
				break;
			case DATASET:
				templateValidator = new ColumnCategoryTemplateValidator(TemplateCategory.DATASET);
				break;
			case GENERIC:
				templateValidator = new ColumnCategoryTemplateValidator(TemplateCategory.GENERIC);
				break;

			default:
				break;
		}

		if(templateValidator!=null){
			HashMap<String, ColumnCategoryConstraint> hash  = templateValidator.getTemplateColumnValidator();
			
			if(hash!=null){
				for (String columnType : hash.keySet()) {
					ColumnCategoryConstraint col = hash.get(columnType);
					violations.add(new ViolationDescription(columnType, col.getConstraintDescription()));
				}
			}
			
		}
		
		return violations;

	}
	
	
	/**
	 * 
	 * @return
	 */
	public static List<String> getOnErrorValues(){
		
		List<String> onErrors = new ArrayList<String>(OnRowErrorAction.values().length);
		for (OnRowErrorAction error : OnRowErrorAction.values()) {
			onErrors.add(error.toString());
		}
		
		logger.info("Returning "+onErrors.size()+" on errors types");
		return onErrors;
	}
	
	/**
	 * 
	 * @return
	 * @throws Exception 
	 */
	public static List<TdTColumnCategory> getTdTColumnCategoryFromTemplateCategory(TemplateCategory template) throws Exception{
		
		if(template==null)
			throw new Exception("TemplateCategory is null");
		
		List<ColumnDescription> listColumns = template.getAllowedColumn();
		if(listColumns==null)
			throw new Exception("ColumnCategory allowed columns is null for template "+template);
		
		
		List<TdTColumnCategory> columns = new ArrayList<TdTColumnCategory>(listColumns.size());
		for (ColumnDescription  col: listColumns) {
			List<Class<? extends DataType>> dataTypes = col.getColumnCategory().getAllowedClasses();
			TdTColumnCategory tdtCC = new TdTColumnCategory(col.getColumnCategory().name(), col.getColumnCategory().name());
			tdtCC.setTdtDataType(getTdTDataTypeFromListDataType(dataTypes));
			columns.add(tdtCC);
		}
		
		logger.info("Returning "+columns.size()+" ColumnCategory");
		
		return columns;
		
	}
	
	public static List<TdTFormatReference> getTdTFormatReferences(Class<? extends DataType> dataType){

		List<ValueFormat> formats = DataTypeFormats.getFormatsPerDataType(dataType);

		List<TdTFormatReference> tdtFormats = new ArrayList<TdTFormatReference>(formats.size());
		for (ValueFormat valueFormat : formats) {
			tdtFormats.add(new TdTFormatReference(valueFormat.getId(), valueFormat.getExample()));
		}
		
		return tdtFormats;
	}
	
	
	/**
	 * 
	 * @return
	 * @throws Exception 
	 */
	public static TdTColumnCategory getTdTColumnCategoryFromColumnCategory(ColumnCategory col) throws Exception{
		
		if(col==null)
			throw new Exception("ColumnCategory is null");
		
//		if(col.isReferenceRequired()){
//			col.getReferenceClass();
//			return new TdTColumnCategory(col.name(), col.name()  )
//		}
		return new TdTColumnCategory(col.name(), col.name());
	}
	

	
	public static List<TdTDataType> getTdTDataTypeFromListDataType(List<Class<? extends DataType>> dataTypes) throws Exception{
		
		List<TdTDataType> listTdTDataType  = new ArrayList<TdTDataType>();
		
		if(dataTypes==null)
			throw new Exception("List of data type is null");
		
		for (Class<? extends DataType> class1 : dataTypes) {
			try{
				listTdTDataType.add(getTdTDataTypeFromDataType(class1));
			}catch (Exception e) {
				logger.error("Errror on converting data type, skypping value",e);
			}
		}
		
		logger.info("Returning "+listTdTDataType.size()+" TdTDataType");
		
		return listTdTDataType;
	}
	
	
	public static TdTDataType getTdTDataTypeFromDataType(Class<? extends DataType> dataType) throws Exception{
		
		if(dataType==null)
			throw new Exception("Data type is null");
		
		String purgedSimpleName = CutStringUtil.stringPurgeSuffix(dataType.getSimpleName(), "Type");
		
		logger.info("\t \t DataType: "+dataType + " purged simple name: " +purgedSimpleName);
		
		List<TdTFormatReference> formats = getTdTFormatReferences(dataType);
		
		if(formats!=null){
			TdTFormatReferenceIndexer indexer = new TdTFormatReferenceIndexer();
			
			for (TdTFormatReference tdTFormatReference : formats) {
				indexer.putFormat(tdTFormatReference);
			}
			
			return new TdTDataType(dataType.getName(), purgedSimpleName, indexer);
		}
		
		return new TdTDataType(dataType.getName(), purgedSimpleName);
	}
	
	public List<String> getOnErrorActions(){
		//TODO
		return null;
	}
	
	public static TemplateUpdaterForDescription getTdTemplateUpdaterFromTemplateDescription(TemplateDescription sdescr, TemplateService service, CacheServerExpressions cache){
		
		if(sdescr==null)
			return null;
		
		logger.info("Converting TemplateDescription...");
		TdTemplateDefinition tdTempDefinition = new TdTemplateDefinition();
		tdTempDefinition.setServerId(sdescr.getId());
		tdTempDefinition.setAgency(sdescr.getAgency());
		tdTempDefinition.setTemplateDescription(sdescr.getDescription());
		tdTempDefinition.setTemplateName(sdescr.getName());
		
		Template stm = sdescr.getTemplate();
		
		logger.info("Converting Template...");
		List<TdColumnDefinition> listTdColumnDef = new ArrayList<TdColumnDefinition>();
		
		if(stm!=null){
			
			//ON ERROR
			if(stm.getOnRowErrorAction()!=null)
				tdTempDefinition.setOnError(stm.getOnRowErrorAction().toString());
			
			//CATEGORY
			TemplateCategory scat = stm.getCategory();
			if(scat!=null){
				tdTempDefinition.setTemplateType(scat.name());
				List<ViolationDescription> constraint = getTemplateCategoryConstraint(scat);
				tdTempDefinition.setTemplateType(new TdTTemplateType(scat.name(), scat.name(), constraint));
			}
			
			List<TemplateColumn<?>> sCol = stm.getColumns();
			
			if(sCol!=null){
				logger.info("Found column size: "+sCol.size());
				for (int i=0; i<sCol.size(); i++) {
					TemplateColumn<?> templateColumn = sCol.get(i);
					
					//ColumnCategory conversion
					ColumnCategory sColCat = templateColumn.getColumnType(); //SERVER 
					TdTColumnCategory tdtCC = new TdTColumnCategory(sColCat.name(), sColCat.name()); //WIDGET
					
					TdTDataType tdDataType = null;
					//Value type conversion
					Class<? extends DataType> sValueType = templateColumn.getValueType(); //SERVER

					try {
						tdDataType = getTdTDataTypeFromDataType(sValueType); //WIDGET
					}catch (Exception e) {
						logger.error("Errror on converting data type, skypping value",e);
					}
					
					//EXPRESSION TODO??
					List<Expression> sExpression = templateColumn.getExpressions();
					List<TemplateExpression> listExpressionExtend = null;
					
					if(sExpression!=null && sExpression.size()>0){
						listExpressionExtend = new ArrayList<TemplateExpression>(sExpression.size());
						
						ExpressionEvaluatorInstance evaluator = new ExpressionEvaluatorInstance(service);
						for (Expression expression : sExpression) {
							
							if(expression!=null){
								logger.info("Found expression : "+expression.toString());
								TemplateExpression cExpression = null;
								try{	
									cExpression = new TemplateExpression();
									cExpression.setServerExpression(expression.toString());
									cExpression.setHumanDescription(evaluator.evaluate(expression));
									listExpressionExtend.add(cExpression);
									logger.info("Updating expression cached: "+expression.toString());
									cache.addExpression(expression);
								}catch (Exception e) {
									logger.error("Expression evalutation exception: ",e);
								}
							}
	//						System.out.println(j +") " +expression.toString());
						}
					}

					TdColumnDefinition tdColumn = new TdColumnDefinition(i, tdtCC, tdDataType, SPECIAL_CATEGORY_TYPE.UNKNOWN);
					tdColumn = settingSpecificReference(templateColumn, tdColumn, service);
					
					if(listExpressionExtend!=null){
						logger.info("Adding expression description, expression size: "+listExpressionExtend.size());
						tdColumn.setRulesExtends(listExpressionExtend);
					}
					listTdColumnDef.add(tdColumn);
				}
			}
		}
		
		TdTemplateUpdater tdUpdater = new TdTemplateUpdater(tdTempDefinition, listTdColumnDef);
		TemplateUpdaterForDescription updaterDescription = new TemplateUpdaterForDescription(tdUpdater, cache);
		logger.info("Returning TemplateUpdaterForDescription "+updaterDescription);
		return updaterDescription;
	}
	
	/**
	 * 
	 * @param templateColumn
	 * @param tdColumn
	 * @param service 
	 * @return
	 */
	public static TdColumnDefinition settingSpecificReference(TemplateColumn<?> templateColumn, TdColumnDefinition tdColumn, TemplateService service){
		
		if(templateColumn==null || templateColumn.getReference()==null)
			return tdColumn;

		Object reference = templateColumn.getReference();
		
		//USED IN CODELIST
		if(reference instanceof LocaleReference){
			logger.info("Found column reference as LocaleReference..");
			LocaleReference locale = (LocaleReference) reference;
			if(locale!=null){
				logger.info("Setting locale as "+locale.getLocale());
				tdColumn.setLocale(locale.getLocale());
//				tdColumn.setSpecialCategoryType(SPECIAL_CATEGORY_TYPE.CODENAME);
			}
		}else if(reference instanceof TimeDimensionReference){
			logger.info("Found column reference as TimeDimensionReference..");
			TimeDimensionReference timeDimension = (TimeDimensionReference) reference;
			if(timeDimension!=null && timeDimension.getPeriod()!=null){

				Map<String, String> timeValues = new HashMap<String, String>();
				
				if(timeDimension.getFormatIdentifier()!=null){
					ValueFormat valueFormat = timeDimension.getPeriod().getTimeFormatById(timeDimension.getFormatIdentifier());
					if(valueFormat!=null)
						timeValues.put(valueFormat.getId(), valueFormat.getExample());
				}
				
				TdTTimePeriod tt = new TdTTimePeriod(timeDimension.getPeriod().name(), timeValues);
				logger.info("Setting TdTTimePeriod as "+tt);
				tdColumn.setTimePeriod(tt);
//				tdColumn.setSpecialCategoryType(SPECIAL_CATEGORY_TYPE.TIMEDIMENSION);
			}
		}else if(reference instanceof FormatReference){
			
			FormatReference formatReference = (FormatReference) reference;
			ValueFormat valueFormat = DataTypeFormats.getFormatPerId(templateColumn.getValueType(), formatReference.getFormatIdentifier());
			tdColumn.getDataType().setReference(new TdTFormatReference(valueFormat.getId(), valueFormat.getExample()));
			
		}else if(reference instanceof DimensionReference){
			logger.info("Found column reference as DimensionReference..");
			DimensionReference dimension = (DimensionReference) reference;
			if(dimension!=null){
				
				String columnId = dimension.getColumnId()!=null?dimension.getColumnId().getValue():null;
				Long tableId = dimension.getTableId()!=null?dimension.getTableId().getValue():null;
				if(columnId!=null && tableId!=null){

					/**
					 * is a codelist so TabResourceType.STANDARD
					 */
					TRId trId = new TRId("", TabResourceType.STANDARD, tableId+"");
					logger.info("Get column for Dimension with TRid "+trId +" and column id: "+columnId);
					ColumnData tempColData = new ColumnData();
					tempColData.setColumnId(columnId);
					tempColData.setTrId(trId);
				
					logger.info("Getting table by table id: "+tableId);
					try{
						Table table = service.getTable(new TableId(tableId));
						if (table.contains(TableDescriptorMetadata.class)){
							TableDescriptorMetadata tdm = table.getMetadata(TableDescriptorMetadata.class);
							if(tdm!=null){
								logger.info("Table name found: "+tdm.getName());
								tempColData.setName(tdm.getName());
							}else
								logger.info("TableDescriptorMetadata for table "+table+" not found");
						}else
							logger.info("Table with id "+tableId+" doesn't contains TableDescriptorMetadata");
						
					}catch (Exception e) {
						logger.warn("Error occurred on recovering table id: "+tableId, e);
					}

					logger.info("Setting Column Data as "+tempColData);
					tdColumn.setColumnData(tempColData);
//					tdColumn.setSpecialCategoryType(SPECIAL_CATEGORY_TYPE.DIMENSION);
				}
			}
		}
		
		return tdColumn;
	}
	
	public static List<String> getAllowedLocales(){
		return new ArrayList<String>(Arrays.asList(Locales.ALLOWED_LOCALES));
	}

	/**
	 * @return
	 */
	public static List<TdTTimePeriod> getTimeDimensionPeriodTypes() {
		
		List<TdTTimePeriod> listTimePeriods = new ArrayList<TdTTimePeriod>(PeriodType.values().length);
		for (PeriodType period : PeriodType.values()){
			
			Map<String, String> valuesFormat = new HashMap<String, String>(period.getAcceptedFormats().size());
			for (ValueFormat vf : period.getAcceptedFormats()) {
				valuesFormat.put(vf.getId(), vf.getExample());
			}
			listTimePeriods.add(new TdTTimePeriod(period.name(),valuesFormat));
		}
		return listTimePeriods;
	}
	
	public static List<TdLicenceModel> getLicences(){
		
		List<TdLicenceModel> licences = new ArrayList<TdLicenceModel>(Licence.values().length);
		for (Licence licence : Licence.values()) {
			licences.add(new TdLicenceModel(licence.toString(), licence.getName()));	
		}
		
		return licences;
	}
	
	public static List<TdBehaviourModel> getDuplicateBehaviours(){

		List<TdBehaviourModel> behs = new ArrayList<TdBehaviourModel>(DuplicateBehaviour.values().length);
		for (DuplicateBehaviour behav : DuplicateBehaviour.values()) {
			behs.add(convertDuplicateBehaviour(behav));	
		}
		
		return behs;
	}
	
	public static TdBehaviourModel convertDuplicateBehaviour(DuplicateBehaviour bhv){

		if(bhv!=null)
			return new TdBehaviourModel(bhv.toString(), bhv.getDescription());
		
		return null;
	}
	
	public static TdFlowModel convertFlow(Collection<TabularResourceMetadata<?>> listMetadata){
		
		TdFlowModel flowModel = new TdFlowModel();
		
		for (TabularResourceMetadata<?> trMetadata : listMetadata) {
			if(trMetadata instanceof NameMetadata){
				flowModel.setName((String) trMetadata.getValue());
			}else if(trMetadata instanceof AgencyMetadata){
				flowModel.setAgency((String) trMetadata.getValue());
			}else if(trMetadata instanceof DescriptionMetadata){
				flowModel.setDescription((String) trMetadata.getValue());
			}else if(trMetadata instanceof RightsMetadata){
				flowModel.setRights((String) trMetadata.getValue());
			}else if(trMetadata instanceof ValidSinceMetadata){
				Calendar calendar = (Calendar) trMetadata.getValue();
				flowModel.setFromDate(calendar.getTime());
			}else if(trMetadata instanceof ValidUntilMetadata){
				Calendar calendar = (Calendar) trMetadata.getValue();
				flowModel.setToDate(calendar.getTime());
			}else if(trMetadata instanceof LicenceMetadata){
				Licence licence = (Licence) trMetadata.getValue();
				flowModel.setLicenceId(licence.getName());
			}
				
		}
		return flowModel;
	}

}
